/* $Id: dliopen.c,v 1.1 89/07/11 03:07:36 dupuy Exp $ */

/*
 * Returns file descriptor for ethernet device by name ("de0", "qe0", etc.).
 * If name is NULL, uses primary ethernet interface.  Will only receive packets
 * of type specified.  Will receive packets for the ethernet address specified,
 * or local ethernet address if NULL.  If there is an error, returns (-1) and
 * the appropriate value is left in errno.  Normal return status zero.
 * Requires superuser privilege.
 */

#include <ctype.h>
#include <errno.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netdnet/dli_var.h>

#include "libether.h"

#ifdef __STDC__

void bcopy (char *from, char *to, int len);

#else

void bcopy ();

#endif

static ether_addr bcast_addr = { { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff } };
static ether_addr ether_multi_addrs[NOFILE];
static struct sockaddr_dl ether_sockaddrs[NOFILE];

#define multi_addr (ether_multi_addrs[fd])
#define sad (ether_sockaddrs[fd])


void
name_to_devid (devid, name)
struct dli_devid *devid;
char *name;
{
    register int i;

    for (i = 0; *name && !isdigit (name[i]) && i < DLI_DEVSIZE; i++)
        devid->dli_devname[i] = name[i];

    devid->dli_devname[i] = '\0';

    devid->dli_devnumber = atoi (&name[i]);
}


void
devid2name (name, devid)
char *name;
struct dli_devid *devid;
{
    register int i;

    for (i = 0; devid->dli_devname[i]; i++)
	name[i] = devid->dli_devname[i];

    name[i++] = '0' + devid->dli_devnumber;

    name[i] = '\0';
}


int
ether_open (name, type, address)
char *name;
int type;
ether_addr *address;
{
    int fd;
    char **interfaces;
    int saved_errno;			/* close seems to blow errno away */

    if (name == 0)
    {					/* get default ethernet interface */
	interfaces = ether_interfaces ();
	if (interfaces == 0 || *interfaces == 0)
	    return (-1);

	name = *interfaces;		/* just use the first name in list */
    }

    if ((fd = socket (AF_DLI, SOCK_DGRAM, DLPROTO_DLI)) < 0)
    {
#ifdef DEBUG
	perror ("ether_open: socket");
#endif
	return (-1);
    }

    if (fd >= NOFILE)			/* not worth making it work here */
    {
        close (fd);
        errno = EMFILE;
	return (-1);
    }

    bzero (&sad, sizeof (sad));
    sad.dli_family = AF_DLI;
    sad.dli_substructype = DLI_ETHERNET;
    name_to_devid (&sad.dli_device, name);
    sad.choose_addr.dli_eaddr.dli_ioctlflg = DLI_EXCLUSIVE;
    sad.choose_addr.dli_eaddr.dli_protype = type;

    if (bind (fd, (struct sockaddr *) &sad, sizeof (sad)) != 0)
    {
	saved_errno = errno;
#ifdef DEBUG
	perror ("ether_open: bind");
#endif
	close (fd);
	errno = saved_errno;
	return (-1);
    }

    if (address != 0)			/* a multicast address */
    {
	if (setsockopt (fd, DLPROTO_DLI, DLI_MULTICAST, (char *) address,
			sizeof (ether_addr)) != 0)
	{
	    saved_errno = errno;
#ifdef DEBUG
	    perror ("ether_open: setsockopt");
#endif
	    close (fd);
	    errno = saved_errno;
	    return (-1);
	}
	multi_addr = *address;
    }
    else
	multi_addr = bcast_addr;

    sad.choose_addr.dli_eaddr.dli_ioctlflg = 0;

    return (fd);
}


/*
 * Reads and returns a single packet, filling in all fields.  If pktbuf is
 * NULL, a buffer is allocated for it.	If pktbuf is not NULL, the function
 * assumes that pktbuf is large enough to hold the packet.  Has to do address
 * checking, since we are operating in DLI_EXCLUSIVE mode.
 */

int
ether_read (fd, packet)
int fd;
ether_packet *packet;
{
    struct sockaddr_dl from;
    int alen;
    int count;
    char *dest;

    if (packet->pktbuf == 0)
    {
	packet->pktlen = ETHER_MAX;
	if ((packet->pktbuf = malloc ((unsigned) ETHER_MAX)) == 0)
	    return (-1);
    }

    from.dli_family = AF_DLI;
    alen = sizeof (from);

    do					/* ignore bogus multicast packets */
    {
	if ((count = recvfrom (fd, packet->pktbuf, (int) packet->pktlen, 0,
			       (struct sockaddr *) &from, &alen)) < 0)
	{
	    if (errno = EWOULDBLOCK)
		errno = EAGAIN;
	    return (-1);
	}

	dest = (char *) from.choose_addr.dli_eaddr.dli_dest;
    }
    while (ETHER_MCAST (dest)
	   && bcmp (dest, (char *) &bcast_addr, sizeof (ether_addr))
	   && bcmp (dest, (char *) &multi_addr, sizeof (ether_addr)));

    bcopy (dest, (char *) packet->dest, sizeof (ether_addr));
    bcopy ((char *) from.choose_addr.dli_eaddr.dli_target,
	   (char *) packet->src, sizeof (ether_addr));
    packet->type[0] = (from.choose_addr.dli_eaddr.dli_protype & 0xff00) >> 8;
    packet->type[1] = from.choose_addr.dli_eaddr.dli_protype & 0xff;

    if (packet->pktlen > count)
	packet->pktlen = count;

    return (count);
}


/*
 * Writes a single packet, filling src and type fields.
 */

int 
ether_write (fd, packet)
int fd;
ether_packet *packet;
{
    bcopy ((char *) packet->dest,
	   (char *) sad.choose_addr.dli_eaddr.dli_target, DLI_EADDRSIZE);

    return (sendto (fd, packet->pktbuf, (int) packet->pktlen, 0,
		    (struct sockaddr *) &sad, sizeof (struct sockaddr_dl)) < 0
	    ? -1 : 0);
}


/*
 * Returns the local ethernet address for the ethernet interface on the file
 * descriptor fd.  The result is stored in the structure given by address; if
 * none is given, malloc is used to allocate space.
 */

ether_addr *ether_address(fd, address)
int fd;
ether_addr *address;
{
    struct ifdevea ifd;

    devid2name (ifd.ifr_name, &sad.dli_device);

    if (ioctl (fd, SIOCRPHYSADDR, (char *) &ifd) < 0)
    {
#ifdef DEBUG
	perror ("ether_addr: ioctl SIOCRPHYSADDR");
#endif
	return (0);
    }

    if (address == 0)
	address = (ether_addr *) malloc (sizeof (ether_addr));

    if (address != 0)
	bcopy ((char *) ifd.current_pa, (char *) address, sizeof (ether_addr));

    return (address);
}
